<?php

/*
+---------------------------------------------------------------------------+
| Max Media Manager v0.3                                                    |
| =================                                                         |
|                                                                           |
| Copyright (c) 2003-2005 m3 Media Services Limited                         |
| For contact details, see: http://www.m3.net/                              |
|                                                                           |
| This program is free software; you can redistribute it and/or modify      |
| it under the terms of the GNU General Public License as published by      |
| the Free Software Foundation; either version 2 of the License, or         |
| (at your option) any later version.                                       |
|                                                                           |
| This program is distributed in the hope that it will be useful,           |
| but WITHOUT ANY WARRANTY; without even the implied warranty of            |
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             |
| GNU General Public License for more details.                              |
|                                                                           |
| You should have received a copy of the GNU General Public License         |
| along with this program; if not, write to the Free Software               |
| Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA |
+---------------------------------------------------------------------------+
$Id: AllocateZoneImpressions.php 3416 2005-06-30 14:40:49Z andrew $
*/

require_once MAX_PATH . '/lib/max/Maintenance/Priority/AdServer/Task.php';
require_once MAX_PATH . '/lib/max/Maintenance/Priority/Constraint.php';
require_once MAX_PATH . '/lib/max/Maintenance/Priority/Entities.php';
require_once MAX_PATH . '/lib/max/Maintenance/Priority/Criteria.php';
require_once MAX_PATH . '/lib/max/Table/Priority.php';
 
/**
 * A class to allocate the required impressions for each advertisement to 
 * zones.
 *
 * Required advertisements impressions are allocated in a set order, so that if
 * there are insufficient impressions available in a zone to satisfy the requirements,
 * more "important" advertisements will have impressions allocated first. The order
 * of advertisement "importance" is:
 *
 * - Advertisements in higher level Priority campaigns are allocated their impressions
 *   before advertisements in lower level Priority campaigns.
 * - Advertisements campaigns with an Expiration Date are allocated their impressions
 *   before advertisements in campaigns without an Expiration Date, given the same
 *   campaign Priority level.
 * - Advertisements in campaigns with an earlier (i.e. sooner) Expiration Date are
 *   allocated their impressions before advertisements in campaigns with a later
 *   Expiration Date, given the same campaign Priority level.
 * - Advertisements with a greater number of (non-infinite) required impressions are
 *   allocated their impressions before advertisements with a smaller number of
 *   (non-infinite) required impressions, given the same campaign Priority level,
 *   and the same campaign Expiration Date or the same lack of Expiration Date.
 *
 * In the above order, required advertisement impressions (obtained from the
 * tmp_ad_required_impression temporary table) are allocated to zones proportionally
 * to the number of remaining (expected) impressions for each possible zone.
 * For example:
 *    
 * - If Advertisement 1 requires 10 impressions, and is linked Zone 1 with 10
 *   expected impressions and Zone 2 with 30 expected impressions, the advertisement
 *   would be allocated (10/40)*10=2.5 impressions in Zone 1, and (10/40)*30=7.5
 *   impressions in Zone 2. This would leave Zone 1 with 7.5 impressions, and Zone 2
 *   with 22.5 impressions.
 * - Then, if Adververtisement 2 requires 10 impressions, and is also linked to 
 *   Zone 1 and Zone 2, then the advertisement would be allocated (10/30)*7.5=2.5
 *   impressions in Zone 1, and (10/30)*22.5=7.5 impressions in Zone 2. This would
 *   leave 5 impressions in Zone 1, and 15 impressions in Zone 2.
 * - Then, if Advertisement 3 requires 50 impressions, and is also linked to Zone
 *   1 and Zone 2, then the advertisements would be allocated (50/20)*5=12.5
 *   impressions in Zone 1 (rounded down to 5 impressions, as 12.5 > 5), and
 *   (50/20)*15=37.5 impressions in Zone 2 (rounded down to 15 impressions, as
 *   37.5 > 15). This would leave no available impressions in Zone 1 and 2 for
 *   any further advertisements.
 *
 * Note: Always do the multiplication before the division in the code, to try to 
 * avoid rounding errors.
 * 
 * @package    MaxMaintenance
 * @subpackage Priority
 * @author     Demian Turner <demian@m3.net>
 * @author     Andrew Hill <andrew@m3.net>
 */
class AllocateZoneImpressions extends MAX_Maintenance_Priority_AdServer_Task
{
    var $aAvailableForecastZoneImpressions;
    var $aAdZoneImpressionAllocations;
    
    /**
     * The constructor method.
     */
    function allocateZoneImpressions()
    {
        parent::MAX_Maintenance_Priority_Adserver_Task();
        $this->table = $this->_getMaxTablePriorityObj();
    }

    /**
     * The main method of the class, that is run by the controlling 
     * task runner class.
     */
    function run()
    {
        // Set the total number of (initially) avaiable zone impressions
        $this->aAvailableForecastZoneImpressions = $this->oDal->getZoneImpressionForecasts();
        // Get the placement data
        $aPlacements = $this->getAllPlacements();
        foreach ($aPlacements as $k => $oPlacement) {
            $aPlacements[$k]->setAdverts();
        	// Populate the ads with data from the
        	// GetRequiredAdImpressions... task job(s) previously run
        	$this->setRequiredImpressions($aPlacements[$k]->adverts);
        	// Order the adverts in the array by the number of impressions
        	// required by each advertisement (greatest first)
        	$this->orderByRequiredImpressions($aPlacements[$k]->adverts);
        }
        // Now that placements have been correctly prioritised, and the
        // advertisements within each placement are ordered correctly too,
        // perform the allocation of advertisement impressions to zones
        $this->aAdZoneImpressionAllocations = array();
        foreach ($aPlacements as $k => $oPlacement) {
    	   $this->allocateImpressions($aPlacements[$k]);
    	}
    	// Save the ad/zone impression allocations to the database
        $this->table->createTable('tmp_ad_zone_impression');
    	$this->oDal->saveAllocatedImpressions($this->aAdZoneImpressionAllocations);
    }
    
    /**
     * A method to get the required impressions for an array of advertisements
     * from the database, and the set these values in the array.
     *
     * @param array $aAdverts An array of Avert objects, passed by reference.
     */
    function setRequiredImpressions(&$aAdverts)
    {
        $aAdvertID = array();
        foreach ($aAdverts as $oAdvert) {
            $aAdvertID[] = $oAdvert->id;
        }
        // Get the required impressions from the database
        $aRequiredImpressions = $this->oDal->getRequiredAdImpressions($aAdvertID);
        // Set the required impressions into the array reference
        foreach ($aAdverts as $k => $oAdvert) {
            if (isset($aRequiredImpressions[$oAdvert->id])) {
                $aAdverts[$k]->requiredImpressions = $aRequiredImpressions[$oAdvert->id];
            } else {
                $aAdverts[$k]->requiredImpressions = 0;
            }
        }
    }
    
    /**
     * A method to order an array of Advert objects in descending order (ie. from
     * largest to smallest) of the requiredImpressions variable of each object.
     *
     * @param array $aAdverts An array of Advert objects, passed by reference.
     */
    function orderByRequiredImpressions(&$aAdverts)
    {
        $impressions = array();
        if (is_array($aAdverts) && (count($aAdverts) > 0)) {
            foreach ($aAdverts as $k => $oAdvert) {
                $impressions[$k] = $oAdvert->requiredImpressions;
            }
            array_multisort($impressions, SORT_DESC, $aAdverts);
        }
    }
    
    /**
     * A method to order an array of zone arrays in descending order (ie. from
     * largest to smallest) of the availableImpressions variable of each object.
     *
     * @param array $aZones An array of zone arrays, passed by reference.
     */
    function orderByAvailableImpressions(&$aZones)
    {
        $impressions = array();
        if (is_array($aZones) && (count($aZones) > 0)) {
            foreach ($aZones as $k => $aZone) {
                $impressions[$k] = $aZone['availableImpressions'];
            }
            array_multisort($impressions, SORT_DESC, $aZones);
        }
    }
    
    /**
     * A method to allocate the impressions of the Adverts in a Placement
     * to the appropriate zone(s), and store the results in the 
     * tmp_ad_zone_impression database table.
     *
     * @param array $oPlacement The placement which required ad impressions allcation.
     */
    function allocateImpressions(&$oPlacement)
    {
        if (is_array($oPlacement->adverts) && (count($oPlacement->adverts) > 0)) {
            // Get the ad/zone associations relevant to this placement
            $aAdvertID = array();
            foreach ($oPlacement->adverts as $oAdvert) {
                $aAdvertID[] = $oAdvert->id;
            }
            $aAdZoneAssociations = $this->oDal->getAdZoneAssociationsByAds($aAdvertID);
            // Iterate through each advertisement in the placement
            foreach ($oPlacement->adverts as $k => $oAdvert) {
                // Set the zone information for the ad, including currently
                // available impressions in each zone
                $oPlacement->adverts[$k]->zones = $aAdZoneAssociations[$oPlacement->adverts[$k]->id];
                // Are there any zones the ad is linked to?
                if (is_array($oPlacement->adverts[$k]->zones) && (count($oPlacement->adverts[$k]->zones) > 0)) {
                    $totalAvaiableImpressions = 0;
                    foreach ($oPlacement->adverts[$k]->zones as $kk => $zone) {
                        // Set the available impressions for this zone (at this point in the
                        // allocation process), and keep track of the total available impressions
                        // in all zones the ad is linked to
                        $oPlacement->adverts[$k]->zones[$kk]['availableImpressions'] = 
                            $this->aAvailableForecastZoneImpressions[$zone['zone_id']];
                        $totalAvaiableImpressions += $oPlacement->adverts[$k]->zones[$kk]['availableImpressions'];
                    }
                    // Order the linked zones in descending order of available impressions
                    $this->orderByAvailableImpressions($oPlacement->adverts[$k]->zones);
                    // Allocate the ad impressions to the associated zones
                    if ($totalAvaiableImpressions > 0) {
                        foreach ($oPlacement->adverts[$k]->zones as $zone) {
                            $allocatedImpressions =
                                $oAdvert->requiredImpressions * 
                                ($zone['availableImpressions'] / $totalAvaiableImpressions);
                            // Is this permitted?
                            if ($allocatedImpressions > $zone['availableImpressions']) {
                                $allocatedImpressions = $zone['availableImpressions'];
                            }
                            // Store the allocated impressions
                            $this->aAdZoneImpressionAllocations[] = array(
                                'ad_id'       => $oAdvert->id,
                                'zone_id'     => $zone['zone_id'],
                                'impressions' => $allocatedImpressions
                            );
                            // Update the available impressions
                            $this->aAvailableForecastZoneImpressions[$zone['zone_id']] -= $allocatedImpressions;
                        }
                    }
                }
            }
        }        
    }
        
   /**
    * Method to return an array of Placement objects, in the following array order:
    *
    * - Placement priority from highest (user defined value > 1) to lowest (1).
    * - Within the same priority level, placements with expiry dates first.
    * - Within placements with expiry dates, placements with sooner expiry dates first.
    *
    * @TODO Does this method also need to order on the following: Within placements
    *       with the same expiry date, and within placements with no expiry date,
    *       placements with the greatest number of remaining required impressions?
    *
    * @return array An array of Placement objects, in the above order.
    */
    function getAllPlacements()
    {
        $orderBys[] = array('priority','DESC');
        $orderBys[] = array('expire','ASC');
        $aPlacements = $this->oDal->getPlacements(array(), array(), array(), $orderBys);
        $aPlacementObjects = array();
        if (is_array($aPlacements)) {
            foreach ($aPlacements as $placement) {
                $aPlacementObjects[] = new Placement($placement);
            }
        }
        return $aPlacementObjects;
    }
     
}

?>
