<?php

/*
+---------------------------------------------------------------------------+
| Max Media Manager v0.3                                                    |
| =================                                                         |
|                                                                           |
| Copyright (c) 2003-2005 m3 Media Services Limited                         |
| For contact details, see: http://www.m3.net/                              |
|                                                                           |
| This program is free software; you can redistribute it and/or modify      |
| it under the terms of the GNU General Public License as published by      |
| the Free Software Foundation; either version 2 of the License, or         |
| (at your option) any later version.                                       |
|                                                                           |
| This program is distributed in the hope that it will be useful,           |
| but WITHOUT ANY WARRANTY; without even the implied warranty of            |
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             |
| GNU General Public License for more details.                              |
|                                                                           |
| You should have received a copy of the GNU General Public License         |
| along with this program; if not, write to the Free Software               |
| Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA |
+---------------------------------------------------------------------------+
$Id: GetRequiredAdImpressionsType1.php 3736 2005-10-10 15:17:56Z andrew $
*/

require_once MAX_PATH . '/lib/max/Maintenance/Priority/AdServer/Task.php';
require_once MAX_PATH . '/lib/max/Maintenance/Priority/Constraint.php';
require_once MAX_PATH . '/lib/max/Maintenance/Priority/Entities.php';
require_once MAX_PATH . '/lib/max/Maintenance/Priority/Criteria.php';
require_once MAX_PATH . '/lib/max/Table/Priority.php';
require_once 'Date.php';

/**
 * A class used to calculate the number of required advertisements in the next
 * operation interval for each ad, such that the ad will meet its delivery
 * requirements.
 *
 * This class is for advertisements in campaigns with Inventory Requirements and
 * a Campaign Expiration Date. The required impressions are calculated using the
 * following algorithm:
 *
 * - If the advertisement is in a campaign that has Click and/or Connection/Conversion
 *   Inventory Requirements, then calculate the campaign's Click Through Ratio and/or
 *   Connection/Conversion Ratio as required, and convert the campaign's remaining
 *   Click and/or Connection/Conversion Inventory Requirements into an equivalent
 *   Impression Inventory Requirement.
 * - Using the smallest of the real or calculated Impression Inventory Requirement values,
 *   distribute the required impressions between the campaign's advertisements, on the 
 *   basis of the campaign's advertisement's weights.
 * - For each advertisement, divide the number of impressions allocated equally between
 *   the available operation intervals remaining in the day(s) before the end of the
 *   campaign. Do not include any operation intervals where the advertisement has been
 *   blocked from delivering (ie. a Day, Time, or Date Delivery Limitation exists). The
 *   number of impressions for each remaining operation interval will be the number of
 *   impressions required for the next operation interval, unless the advertisement is
 *   blocked from delivering in the next operation interval, in which case the required
 *   impressions for the next operation interval will be zero.
 *
 * @package    MaxMaintenance
 * @subpackage Priority
 * @author     Demian Turner <demian@m3.net>
 * @author     James Floyd <james@m3.net>
 * @todo       Implement error handling
 */
class GetRequiredAdImpressionsType1 extends MAX_Maintenance_Priority_AdServer_Task
{
    
    var $table;
      
    function getRequiredAdImpressionsType1()
    {
        parent::MAX_Maintenance_Priority_Adserver_Task();
        $this->table = $this->_getMaxTablePriorityObj();
    }

    /**
     * The main method of the class, that is run by the controlling 
     * task runner class.
     */
    function run()
    {
        $aAllPlacements = $this->getValidPlacements();
        if (is_array($aAllPlacements) && (count($aAllPlacements) > 0)) {
            foreach ($aAllPlacements as $k => $oPlacement) {
                $this->getPlacementImpressionInventoryRequirement($aAllPlacements[$k]);
                $aAllPlacements[$k]->setAdverts();
            }
            $this->distributePlacementImpressions($aAllPlacements);
        }
    }
    
    /**
     * A method that uses the getAllPlacements() method to obtain all placements
     * that meet the requirements of this class. That is:
     *
     * - The placement has an expiration date (that is not already passed); and
     * - As a result of the above, the placement must have an activation date that has
     *   passed, or has the default fake activation date; and
     * - The placement is active, and has a priority of at least 1; and
     * - The placement has inventory requirements for the duration of its activation.
     *
     * @return array An array of Placement objects.
     */
    function getValidPlacements()
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        // Get current date
        $oDate = $this->_getDate();
        $dateYMD = $oDate->getYear() . '-' . $oDate->getMonth() . '-' . $oDate->getDay();
        $table = $conf['table']['prefix'] . $conf['table']['campaigns'];
        $wheres = array(
                        array("($table.activate = '0000-00-00' OR $table.activate <= '$dateYMD')", 'AND'),
                        array("$table.expire >= '$dateYMD'", 'AND'),
                        array("$table.priority >= 1", 'AND'),
                        array("$table.active = 't'", 'AND'),
                        array("($table.views > 0 OR $table.clicks > 0 OR $table.conversions > 0)", 'AND'),
                        );
        return $this->getAllPlacements(array(), $wheres);
    }
    
    /**
     * A method to return the current date. Can return the current date:
     *
     * - Based on the optional parameter; or
     * - From the current date registered with the Service Locator; or
     * - Based on the time when called.
     *
     * @access private
     * @param string $date An optional string representation of the current date.
     * @return object The current date.
     */
    function _getDate($date = '')
    {
        if (!empty($date)) {
            return new Date($date);
        }
        $oServiceLocator = &ServiceLocator::instance();
        $oDateNow = &$oServiceLocator->get('now');
        if (!$oDateNow) {
            $oDateNow = &new Date();
            $oServiceLocator->register('now', $oDate);
        }
        return $oDateNow;
    }
    
    /**
     * Method to return an array of Placement objects, given a set of filters.
     *
     * Essentially a convenience method to convert the results of the 
     * {@link MAX_Dal_Maintenance::getPlacements()} method from an array of
     * database records into an array of Placement objects.
     *
     * @param array $fields An optional array of extra fields to select from the database
     *                      (see the {@link MAX_Dal_Maintenance::getPlacements()} class.)
     * @param array $wheres An optional array of where statements to limit which placements
     *                      are returned from the database (see the
     *                      {@link MAX_Dal_Maintenance::getPlacements()} class.)
     * @return array An array of Placement objects, appropriate to the $filter given.
     */
    function getAllPlacements($fields = array(), $wheres = array())
    {
        $aPlacementObjects = array();
        $aPlacements = $this->oDal->getPlacements($fields, $wheres);
        if (is_array($aPlacements) && (count($aPlacements) > 0)) {
            foreach ($aPlacements as $placement) {
                $aPlacementObjects[] = new Placement($placement);
            }
        }
        return $aPlacementObjects;
    }
    
    /**
     * Method to estimate the impressions required to fulfill a given
     * placement lifetime impression, click, or conversion requirement. If
     * more than one requirement exists the smallest calculated impression
     * requirement will be returned.
     *
     * The $oPlacement parameter is passed by reference and will have
     * the calculated impression requirement added to it in the position
     * $oPlacement->requiredImpressions
     * 
     * @param object $oPlacement
     * @return void
     */
    function getPlacementImpressionInventoryRequirement(&$oPlacement)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        // Get campaign summary statistic totals
        $oPlacement->setSummaryStatisticsToDate();
        // Calculate impressions required to fulfill click requirement
        $clickImpressions = $this->getInventoryImpressionsRequired(
            $oPlacement->clicks,
            $conf['priority']['defaultClickRatio'],
            $oPlacement->sumClicks,
            $oPlacement->sumViews);
        // Calculate impressions required to fulfill conversion requirement
        $conversionImpressions = $this->getInventoryImpressionsRequired(
            $oPlacement->conversions,
            $conf['priority']['defaultConversionRatio'],
            $oPlacement->sumConversions,
            $oPlacement->sumViews);
        // Get impression requirement
        if ($oPlacement->views > 0) {
            $impressions = $oPlacement->views - $oPlacement->sumViews;
        } else {
            $impressions = 0;
        }
        // Choose smallest required impression
        $oPlacement->requiredImpressions = $this->getSmallestNonZeroConstraint(
            array(
                $clickImpressions, 
                $conversionImpressions, 
                $impressions
            )
        );
    }
    
    /**
     * Method to calculate the number of impressions required to achieve delivery 
     * of a given click or conversion inventory.
     *
     * If historical delivery data does not exist, the ratio provided in $defaultRatio
     * will be use for all calculations.  Please be aware that a default a click ratio
     * is provided in the system wide configuration file and should be a value between
     * zero and one ((ratio > 0) && (ratio <= 1)).
     *
     * $GLOBALS['_MAX']['CONF']['priority']['defaultClickRatio']
     * $GLOBALS['_MAX']['CONF']['priority']['defaultConversionRatio']
     *
     * @param integer $inventory Total placement inventory to be achieved over lifetime/day.
     * @param integer $defaultRatio Click/conversion ratio used if historical data does not exist.
     * @param integer $inventoryToDate Total clicks/conversions achieved to date over lifetime/day.
     * @param integer $impressionsToDate Total impressions delivered to date over lifetime/day.
     * @return integer
     */
    function getInventoryImpressionsRequired($inventory, $defaultRatio, $inventoryToDate = 0, $impressionsToDate = 0)
    {
        $requiredImpressions = 0;
        if ($inventory > 0) {
            // if historical information exists
            if ($inventoryToDate > 0) {
                $inventoryPerImpression = ($inventoryToDate / $impressionsToDate);
            } else {
                $inventoryPerImpression = $defaultRatio;
            }
            $requiredImpressions = (int)$inventory / $inventoryPerImpression;
        }
        return ceil($requiredImpressions);
    }
    
    /**
     * Returns the smallest non zero value of a given array of values.
     *
     * If $aValues is empty then zero is returned
     *
     * @param array $aValues An array of values to evaluated.
     * @return integer The smallest non-zero value from the array, or zero if empty.
     */
    function getSmallestNonZeroConstraint($aValues = array())
    {
        if (!is_array($aValues) || (count($aValues) == 0)) return 0;
        foreach ($aValues as $val) {
            if ((($val > 0) && !isset($minVal)) ||
                (($val > 0) && isset($minVal) && ($val < $minVal))) {
                $minVal = $val;
            }
        }
        return (isset($minVal)) ? $minVal : 0;
    }
    
    /**
     * A method to obtain a new {@link MAX_Maintenance_Priority_Criteria}
     * object.
     *
     * @access private
     */
    function _getMaxMaintenancePriorityCriteria($aConstraints)
    {
        return new MAX_Maintenance_Priority_Criteria($aConstraints);
    }
    
    /**
     * Distribute the calculated required placement impressions between 
     * associated child ads. Impression allocation takes in to account 
     * ad weight, the number of operations intervals the ad will active 
     * in given date/time delivery constraints.
     *
     * The calculated ad impressions are written to the temporary table 
     * tmp_ad_required_impression for later analysis by the 
     * {@link AllocateZoneImpressions} class.
     *
     * @param array $aPlacements An array of {@link Placement} objects which required that the
     *                           total required impressions be distributed between the component
     *                           advertisements.
     *
     * @TODO Update code to calculate required impressions for each ad for one week in advance,
     *       not just the next operation interval.
     * @TODO Make sure impressions are being distributed as accurately as possible.
     * @TODO Check that campaign has delivery requirments $aCampaign['inventoryImpressionReq'] (?)
     * @TODO Minimise rounding errors.
     */
    function distributePlacementImpressions($aPlacements)
    {
        $conf = $GLOBALS['_MAX']['CONF'];
        // Create an array for storing required ad impressions
        $aRequiredAdImpressions = array();
        // Get the current operation interval start/end
        $aCurrentOperationIntervalDates =
            $this->oOperationInt->convertDateToOperationIntervalStartAndEndDates($this->_getDate());
        // For each placement
        foreach ($aPlacements as $k => $oPlacement) {
            // Get date object to represent placement expire date
            if ($oPlacement->expire == '0000-00-00' &&
                (($oPlacement->target_impression > 0) ||
                 ($oPlacement->target_click > 0) ||
                 ($oPlacement->target_conversion > 0))) {
                // The campaign has a daily target, so treat this placement
                // as if it ends at the end of today
                $oDate = &$this->_getDate();
                // Get the end of the day from this date
                $placementExpiryDate = &new Date($oDate->format('%Y-%m-%d') . ' 23:59:59');
            } else {
                $placementExpiryDate = &$this->_getDate($oPlacement->expire);
                // Placement expires at end of expiry date so add one day - one
                // second so we have a date with time portion 23:59:59
                $placementExpiryDate->addSeconds(SECONDS_PER_DAY - 1);
            }
            // Determine number of remaining operation intervals for placement
            $placementRemainingOperationIntervals =
                $this->oOperationInt->getRemaining(
                    $aCurrentOperationIntervalDates['start'],
                    $placementExpiryDate
                );
            // Sum the weights of all ads in placement
            $totalWeight = $this->getPlacementAdWeightTotal($oPlacement);
            if (PEAR::isError($totalWeight, MAX_ERROR_INVALIDARGS)) {
                /**
                 * @TODO Ensure that this PEAR::Error is handled by calling code, or 
                 *       raise an error instead of returning.
                 */
                return $totalWeight;
            }
            // Calculate number impressions per weight value of 1 (one)
            $totalImpressionsPerUnitWeight = $oPlacement->requiredImpressions / $totalWeight;
            // For each advertisement
            foreach ($oPlacement->adverts as $kk => $oAdvert) {
                // Get impressions required, based on the ad weight
                $requiredAdImpressions = 0;
                if ($oAdvert->weight > 0) {
                    $requiredAdImpressions = $oAdvert->weight * $totalImpressionsPerUnitWeight;
                }
                if ($requiredAdImpressions > 0) {
                    /**
                     * @TODO check if constraints exist.
                     */
                    $aConstraints = $oAdvert->getConstraints();
                    $criteria = $this->_getMaxMaintenancePriorityCriteria($aConstraints);
                    if ($criteria->isLimitedNow($aCurrentOperationIntervalDates) === false) {
                        // Find number of active operation intervals 
                        $activeAdOpInts =
                            $criteria->getActiveAdOperationIntervals(
                                $placementRemainingOperationIntervals,
                                $aCurrentOperationIntervalDates,
                                $placementExpiryDate
                            );
                        // Are there active intervals for the ad?
                        if ($activeAdOpInts > 0) {
                            // Divide number required impressions between active operation intervals
                            /**
                             * @TODO Decide which of round(), ceil() or floor() is correct logic.
                             */
                            $oAdvert->setRequiredImpressions(
                                round($requiredAdImpressions / $activeAdOpInts)
                            );
                            
                            // Delivery hack
                            if ($conf['maintenance']['deliveryHack']) {
                                $oServiceLocator = &ServiceLocator::instance();
                                $oDateNow = &$oServiceLocator->get('now');
                                $hour = $oDateNow->getHour();
                                $factor = 0.18;
                                switch ($hour) {
                                    case 8:
                                        $factor = 0.75;
                                        break;
                                    case 9:
                                        $factor = 1.19;
                                        break;
                                    case 10:
                                        $factor = 1.48;
                                        break;
                                    case 11:
                                        $factor = 1.55;
                                        break;
                                    case 12:
                                        $factor = 1.96;
                                        break;
                                    case 13:
                                        $factor = 1.83;
                                        break;
                                    case 14:
                                        $factor = 1.38;
                                        break;
                                    case 15:
                                        $factor = 1.53;
                                        break;
                                    case 16:
                                        $factor = 1.55;
                                        break;
                                    case 17:
                                        $factor = 1.56;
                                        break;
                                    case 18:
                                        $factor = 1.73;
                                        break;
                                    case 19:
                                        $factor = 1.71;
                                        break;
                                    case 20:
                                        $factor = 1.27;
                                        break;
                                    case 21:
                                        $factor = 1.19;
                                        break;
                                    case 22:
                                        $factor = 1.18;
                                        break;
                                    case 23:
                                        $factor = 0.72;
                                        break;
                                }
                                $oAdvert->setRequiredImpressions(
                                    round(($requiredAdImpressions / $activeAdOpInts) * $factor)
                                );
                            }
                            
                            
                            $aRequiredAdImpressions[] = array(
                                'ad_id'                => $oAdvert->id,
                                'required_impressions' => $oAdvert->requiredImpressions
                            );
                        }
                    }
                }
            }
        }
        // Save the required impressions into the temporary database table
        $this->table->createTable('tmp_ad_required_impression');
        $this->oDal->saveRequiredAdImpressions($aRequiredAdImpressions);
    }
    
     /**
     * Method to sum all ad weight values for a given placement. If no 
     * ads, or if the sum is zero, the default summation weight is set
     * to unity (1).
     *
     * @param object $oPlacement A Placement object.
     * @return integer The ad weight total.
     */
    function getPlacementAdWeightTotal($oPlacement)
    {
        $weight = 0;
        if (is_a($oPlacement, 'Placement')) {
            foreach ($oPlacement->adverts as $oAdvert) {
                $weight += ($oAdvert->weight > 0) ? $oAdvert->weight : 0;
            }
            if ($weight === 0) {
                $weight = 1;
            }
            return $weight;
        }
        return PEAR::raiseError('Invalid Placement object argument', MAX_ERROR_INVALIDARGS);
    }
    
}

?>
